/* The chess evaluator.  Compatible to `etc/CHESS-EVALUATOR'. 
 */

#include "opt_main.hh"
#include "string.hh"
#include "position.hh"
#include "io.hh"
#include "score.hh"
#include "generator.hh"
#include "error.hh"
#include "material.hh"
#include "main.hh"
#include "time.hh"
#include "read_line.hh"
#include "exit.hh"
#include "square_value.hh"
//#include "scan.hh"
#include "read_fen.hh"
#include "write_formatted.hh"

bool option_verbose;
bool option_interactive; 

#define Write_Verbose(args...)			\
{						\
	if (option_verbose)			\
	{					\
		Write_Formatted(IO::err,	\
			    args);		\
	}					\
}

static const int piece_value_score[6]= 
{100,
 300,
 300,
 450,
 900,
 0};

Inline
int 
Piece_Value(Piece_Type pt)
{
	Assert (pt >= pt_P && pt <= pt_K);

	return piece_value_score[pt];
}

int 
Square_Score(int square,
	     Piece_Type pt,
	     const Material &material)
{
	int result= 0;

	switch (pt)
	{
	default:
	{
		const int val[8]= {1, 10, 13, 20, 20, 13, 10, 1};
		int score= val[Square_File(square)] * val[Square_Rank(square)] / 10;
		if (pt == pt_K &&
		    material(pt_Q, cc_w) != 0 &&
		    material(pt_Q, cc_b) != 0)
			score= - score;
		if (pt == pt_P)
			score *= 2;
		result += score;
	}

	case pt_Q:
		break;

	case pt_P:
		result += Square_Value(square, sv_pawn);
		break;
	}

	return result;
}

int 
Piece_Score(int square,
	    Piece p,
	    const Material &material)
{
	return (Piece_Value(p.Get_Type()) + 
		Square_Score(p.Get_Color() == cc_w ? square : Square_Mirror(square),
			    p.Get_Type(), material));
}

int
Move_Score(const Move &move,
	  Piece piece,
	   const Material &material)
	/* How MOVE changes the position.  PIECE is the moving piece.
	   COUNT is the piece count before the move. 
	 */
{
	int result= 0;
	if (move.captured_piece.Is())
		result += Piece_Score(move.Get_Captured_Square(), 
				      move.captured_piece, 
				      material);
	if (move.promotion.Is())
	{
		result += Piece_Score(move.to, move.promotion, material);
		result -= Piece_Score(move.from, piece, material);
	}
	else
	{
		bool black= piece.Get_Color() == cc_b;
		result += Square_Score(black ? Square_Mirror(move.to) : move.to,
				       piece.Get_Type(), 
				       material);
		result -= Square_Score(black ? Square_Mirror(move.from) : move.from,
				       piece.Get_Type(), 
				       material);
		if (move.castling != Move::no_castling)
		{
			result += Square_Score(black
					      ? Square_Mirror(move.Get_Rook_To()) 
					      : move.Get_Rook_To(),
					       pt_R, material);
			result -= Square_Score(black
					      ? Square_Mirror(move.Get_Rook_From())
					      : move.Get_Rook_From(),
					 pt_R, material);
		}
	}

	return result;
}

int
Position_Score(const Position &position,
	       const Material &material)
	/* From white's viewpoint. 
	 */
{
	int result= score_even;
	
	for (int file= sq_a; file <= sq_h; ++ file)
		for (int rank= sq_1; rank <= sq_8; ++ rank)
		{
			Piece c= position.Get_Piece(Make_Square(file, rank));
			if (c.Is())
			{
				int val= Piece_Score(Make_Square(file, rank),
						    c,
						     material);
				if (c.Get_Color() == cc_w)
					result += val;
				else 
					result -= val;
				
			}
		}

	return result;
}

class Score_Position
/* A position that keeps track of:
 *         - an evaluation of the position (always from the side to move)
 */
	:  public Position
{
public:

	Score_Position(const Position &position)
		:  Position(position)
		{
#if 0
			for (int pt= 0;  pt < 6;  ++pt)
				for (int cc= 0;  cc < 2;  ++cc)
					material(pt, cc)= 0;
			for (int file= sq_a;  file <= sq_h;  ++file)
				for (int rank= sq_1;  rank <= sq_8;  ++rank)
				{
					Piece piece= Get_Piece(Make_Square(file, rank));
					if (piece.Is())
						++ material(piece.Get_Type(),
							    piece.Get_Color());
				}
#endif /* 0 */ 
			score= Position_Score((Position &)This, Get_Material());
			if (position.Get_To_Move() == cc_b)
				score= - score;
		}

	int Get_Score() const
		{
			return score;
		}

	void operator << (const Move &move)
		{
			score += Move_Score(move, Get_Piece(move.from), Get_Material());
			score= - score;
#if 0
			if (move.captured_piece.Is())
				-- material(move.captured_piece.Get_Type(), CC_Turn(to_move));
#endif /* 0 */ 
			Position::operator << (move);
		}

	void operator >> (const Move &move)
		{
			Position::operator >> (move);
#if 0
			if (move.captured_piece.Is())
				++ count(move.captured_piece.Get_Type(), CC_Turn(to_move));
#endif /* 0 */ 
			score= - score;
			score -= Move_Score(move, Get_Piece(move.from), Get_Material());
		}


#if 0
	Material &Get_Count()
		{
			return count;
		}
#endif /* 0 */ 

private:

	int score;
	/* From the viewpoint of the side to move. 
	 */
#if 0
	Material count;
#endif /* 0 */
};

bool
Quiet(const Move &m)
{
	return m.captured_piece.Is_Empty() && m.promotion.Is_Empty();
}

void
Evaluate(Score_Position &position,
	 int depth,
	 int &score,
	 int &index,
	 int min,
	 int max)
/* Evaluate POSITION.  Calculate DEPTH.  Write evaluation into SCORE.
   If there are legal moves, write the best into INDEX, else set INDEX
   to -1.  Calculate position from viewpoint of side to move. 
 */
{
	Assert (depth >= -1);

	if (option_interactive)
	{
		Write_Formatted(IO::out, " > ");
		IO::out.Flush();
		String line;
		if (Read_Line(IO::in, line))
		{
			if (line == "quit")
			{
				Exit(1);
			}
		}
	}

	static int knots= 0;
	++ knots;

	static int interval= 128; /* ms */

	if (knots % interval == 0)
	{
		static Date last_date= DATE_NO_DATE;
		Date now= Get_Date();
		int elapsed= Date_Is(last_date) ? 0 >? Date_Diff_ms(last_date, now) : -1;

		if (elapsed != -1 && elapsed < 1000)
		{
			interval *= 2;
		}
		Write_Formatted(IO::out,
			    "knots %d\n",
			    knots);
		last_date= now;
	}

	Generator generator;
	generator.New(position);
	int count= *generator;

	if (count == 0)
	{
		index= -1;
		score= position.Is_Check() ? score_is_mate : score_even;
		return;
	}

	if (depth <= 0)
	{
		bool cont= false;
		
		if (depth == 0)
			for (int i= 0; i < count && ! cont; ++i)
				if (! Quiet(generator.Get_Move(i)))
					cont= true;

		if (! cont)
		{
			score= position.Get_Score();
			if (option_verbose)
			{
				Write_Formatted(IO::err,
					    "\t\tposition is quiet.\n"
					    "\t\t[%d]\n",
					    score); 
			}
			index= 0;
			return;
		}
	}

	Assert (depth >= 0);

	int order[count];
	int values[count];
	/* Real indexes. */

	for (int i= 0; i < count; ++i)
		order[i]= i;

	for (int d= -1;  d <= (depth - 1 >? 0);  ++d)
	{
		int bestval= score_is_mate;
		for (int i= 0; i < count; ++i)
		{
			if (d == -1)
			{
				Move m= generator.Get_Move(order[i]);
				values[order[i]]= 
					(position.Get_Score()                +
					 Move_Score(m, 
						   position.Get_Piece(m.from),
						    position.Get_Material())    +
					 count);
			}
			else
			{
				int old_score= position.Get_Score();
				if (option_verbose)
				{
					Write_Formatted(IO::err,
						    "\t\t(%d)\n",
						    position.Get_Score()); 
				}
				position << generator.Get_Move(order[i]);

				if (option_verbose)
				{
					String move= Format_Move(generator[order[i]]); 
					Write_Formatted(IO::err,
						    "\t-->  %s\n"
						    "\t\t(%d)\n",
						    move(),
						    position.Get_Score()); 
				}

				int val;
				int ind;

				Evaluate(position, 
					  d,
					  val,
					  ind,
					  - max,
					  - bestval <? - min);

				if (option_verbose)
				{
					String move= Format_Move(generator[order[i]]); 
					Write_Formatted(IO::err,
						    "\t\t[%d]\n"
						    "\t<--  %s\n",
						    val,
						    move());
				}

				position >> generator.Get_Move(order[i]);
				Assert (old_score == position.Get_Score()); 

				if (depth == 0 && d == 0)
				{
					Assert (i == 0);
					int last= Score_Negate(val);
					int current= position.Get_Score(); 
					Write_Verbose("\t\tlast=    %d, current= %d\n",
						      last, current);

					if (last > current)
					{
						score= last;
						Write_Verbose("\t\tassuming last move is best.\n");
					}
					else
					{
						score= current;
						Write_Verbose("\t\tlast move makes position worse.\n");
					}
					index= order[0];

					return; 
				}

				int newval= Score_Negate(val);
				Write_Verbose("\t\tval:  lastmove= %d, best= %d\n",
					      newval, bestval);
				if (newval > bestval)
				{
					Write_Verbose("\t\tnew best move.\n");
					bestval= newval;
				}
				else
				{
				}
				values[order[i]]= newval;
				Write_Verbose("\t\tcutoff:  lastmove= %d, max= %d\n", newval, max);
				if (newval >= max)
				{
					Write_Verbose("\t\tcutting off.\n");
					for (int j= i + 1; j < count; ++j)
					{
						values[order[j]]= newval;
					}
					i= count;
				}
			}
		}
		
		for (int i= 0; i + 1 < count; ++i)
		{
			int best= i;
			for (int j= i + 1; j < count; ++j)
			{
				if (values[order[j]] > values[order[best]])
					best= j;
			}
			if (best != i)
			{
				int tmp= order[i];
				order[i]= order[best];
				order[best]= tmp;
			}
		}
	}

	index= order[0];
	score= values[index];

}

const char *opt_main_options= "p:vid:";

int
Opt_Main(int argc,
	 char **argv)
{
	String fen= Get_Opt_Arg('p');
	String max_depth_text= Get_Opt_Arg('d');
	int max_depth= -1;
	
	if (max_depth_text.Is())
	{
		if (1 != sscanf(max_depth_text(),
			      "%d",
			      & max_depth))
		{
			Write_Formatted(IO::err,
				    "%s:  *** Error:  invalid argument `%s' to `-d'.\n",
				    Main_Get_Name()(),
				    max_depth_text());
			return 1;
		}
	}

	option_verbose= Get_Opt('v'); 
	option_interactive= Get_Opt('i');

	if (option_verbose)
	{
		Write_Formatted(IO::err,
			    "\tVerbose mode output:\n"
			    "\t\t(SCORE)     Value of position from the perspective\n"
			    "\t\t           of the side to move.\n"
			    "\t\t[SCORE]     Evaluation of position after calculation.\n"
			    "\t\t--> MOVE   MOVE is played.\n"
			    "\t\t<-- MOVE   MOVE is taken back.\n"
			    "\n");
	}

	if (option_interactive)
	{
		Write_Formatted(IO::err,
			    "\tInteractive mode:  Enter `quit' to exit.\n"
			    "\n");
	}

	for (int i= 0;  i < argc;  ++i)
	{
		fen += argv[i];
		if (i + 1 < argc)
			fen += ' ';
	}

	Position position;

	if (fen.Is())
	{
		if (! Read_FEN(fen,
			       position))
		{
			String error= Error::Get();
			Write_Formatted(IO::err,
				    "%s:  *** Error:  %s:  invalid FEN position:  %s\n",
				    Main_Get_Name()(), 
				    fen(),
				    error());
			return 1;
		}
	}
	else
		position.Reset();

	Generator generator;
	generator.New(position);

	if (*generator == 0)
	{
		if (position.Is_Check())
			Write_Formatted(IO::out,
//				    "mate\n"
				    "score %%0\n");
		else
			Write_Formatted(IO::out,
//				    "stalemate\n"
				    "score 0-100\n");
		return 0;
	}

	Score_Position score_position(position);

	for (int depth= 1;  depth != (max_depth + 1);  ++ depth)
	{
		Write_Formatted(IO::out,
			    "depth %d\n",
			    depth);

		int score;
		int index;

		Evaluate(score_position, 
			 depth,
			 score,
			 index,
			 score_is_mate,
			 score_mate_in_1);

		String move= generator.Get_Simple_Move_Name(index);

		Write_Formatted(IO::out,
			    "main %s\n",
			    move());

		String evaluation= Format_Score(score);

		Write_Formatted(IO::out,
			    "move %s %s\n",
			    move(),
			    evaluation());
		
		String ev= Format_Score(score);
		Write_Formatted(IO::out,
			    "score %s\n",
			    ev());

		if (score > score_not_win    ||
		    score < score_not_lose)
			return 0;
	}

	return 0; 
}

const char *
arg_get_help=
"The BabyChess evaluator.\n"
		"\n"
		"Analyze a chess position, while writing thinking information\n"
		"on the output.  Stop as soon as a mate is found.\n"
		"\n"
		"Options:\n"
		"\t-p \"POSITION\"    The position to analyze, in FEN notation.\n"
		"\t                 If not specified, use the standard initial\n"
		"\t                 position.\n"
		"\t-v               Verbose mode.  Write analyze information to\n"
		"\t                 error output.\n"
		"\t-i               Interactive mode.  Read one line of input\n"
		"\t                 before each step.\n"
		"\t-d DEPTH         Calculate up to depth DEPTH.  A value of -1\n"
		"\t                 indicates to not stop.  Default is -1.\n"; 
