#ifndef EDITOR_WINDOW_INCLUDED
#define EDITOR_WINDOW_INCLUDED

/* A window that edits an object.  Editor_Window is a base class of: 

   - Object_Editor:  Object that can be edited but not saved to file,
                     e.g. board settings. 

   - Document_Editor:  Objects that are written to files, e.g. Game. 		    
 */

#include "buffer.hh"
#include "packer.hh"
#include "error.hh"

enum {use_ok,     /* OK                                  */
      use_no,     /* Finish editing, don't save changes  */
      use_retry}; /* Continue editing                    */

template <class Type>
class Editor_Window
	:  /* virtual public Popup_Window, */
	   public V_Packer
{
public:

	typedef Type Edited;

	virtual const Type &Editor_Window_Get_Value() = 0;
	/* May be called after widget was closed. 
	 */
	virtual bool Editor_Window_Transient() const
		{
			return true; 
		}
	virtual const char *EW_Type_Name() const
		/* The full type name, all words capitalized. 
		 */
		{
			return "Type";
		}
	virtual const char *EW_Short_Type_Name() const
		/* Type name in one word, beginning with an upper case
		   letter. 
		 */
		{
//			Assert_Unreached(); 
			return ""; 
		}

	void MR_Receive(int msg);
	void Wdg_Close()
		{
			V_Packer::Wdg_Close();
		}
	String Wdg_Init_Caption() const
		{
			return EW_Type_Name(); 
//			return Trl("Notation settings");
		}

	bool Open(Pointer <Widget> parent,
		  int new_msg_edited= -1,
		  Pointer <Widget> owner= NULL, 
		  Pointer <const Type> initial_value= NULL,
		  Pointer <Widget> parent_window= NULL,
		  bool ignore_value= false);
	/* If OWNER is not NULL, MSG_EDITED is sent to OWNER when
	   editing is finished.  If IGNORE_VALUE is TRUE,
	   Editor_Window_Set_Value() must be called before Open(). 
	*/

	bool Close_Editor(bool force)
		/* Return whether was closed. 
		 */
		{
			if (must_close || force)
			{
				V_Packer::Close(true);
				/* Force closing. 
				 */
				owner= NULL; 
				parent_window= NULL; 
				
				return true;
			}
			else
				return false;
		}
	Pointer <Widget> Get_Owner() const
		{
			return owner;
		}
protected:

	virtual bool Editor_Window_Open(Pointer <Packer> parent)= 0;
	/* Open the editor widget.  Set_Value() has already been called at
	   this moment.  Append widgets to PARENT which is a V_Packer. 
	*/
    
	virtual int Editor_Window_Use() {return use_ok; }
	/* Called before the value is used.  Return one of the use_* constants. 
	 */

	virtual void Editor_Window_Set_Value(Pointer <const Type> value)= 0;
	/* Initialize editor with VALUE.  The window is not open at
	   that moment.  If VALUE is NULL, set to default value.
	   Must be called before Open() when IGNORE_VALUE is set in
	   Open(). 
	*/
	
	virtual bool Editor_Window_Change_Value(const Type &) 
		/* Change value when editor is open.  Return TRUE on success.
		   Not implemented by default.  
		*/
		{
			Error::Set_No_Sys();
			return false; 
		}

	virtual bool Editor_Window_New()
		/* Make new when window is open. 
		 */
		{
			Error::Set_No_Sys();
			return false;
		}

	virtual bool Editor_Window_Try_Load(String /* file_name */ ,
					    Pointer <Type> /* object */ ) 
		/* Try to load FILE, which is not of type Type.  Return
		   TRUE on success.  Fill OBJECT with content of
		   FILE_NAME. 
		*/
		{ 
			return false;  
		}

private:
    
	int msg_edited;
	Pointer <Widget> owner;
	Pointer <Widget> parent_window; 

	bool must_close;
	/* Set before MSG_EDITED is sent to owner.  TRUE when window
	   must be closed after the owner read the value. 
	*/

	virtual void Editor_Window_Apply()
		/* When the value is applied. 
		 */
		{
		}
};

template <class Type>
bool Editor_Window <Type> ::
Open(Pointer <Widget> parent,
     int new_msg_edited,
     Pointer <Widget> new_owner, 
     Pointer <const Type> initial_value,
     Pointer <Widget> new_parent_window,
     bool ignore_value)
{
	if (! ignore_value)
		Editor_Window_Set_Value(initial_value);
	msg_edited= new_msg_edited;

	if (! V_Packer::Open(parent))
		return false;

	owner= new_owner;
	parent_window= new_parent_window;
	if (! parent_window.Is())
		parent_window= owner; 

	if (! Editor_Window_Open(this))
	{
		V_Packer::Close(true);
		return false; 
	}

	if (owner.Is() && Editor_Window_Transient())
		Set_Transient(parent_window()); 

	return true;
}

template <class Type>
void Editor_Window <Type> ::MR_Receive(int msg)
{
	switch (msg)
	{
			case msg_cancel:
				Close();
				break;

	case msg_ok:
		switch (Editor_Window_Use())
		{
		case use_ok:
			must_close= true;
			if (owner.Is())
			{
				owner().MR_Send(msg_edited); 
				parent_window().Activate_Window();
			}

			/* fall through */
		case use_no:
			Close();
			break;

		case use_retry:
			break;
		}
		break;

	case msg_apply:
		switch (Editor_Window_Use())
		{
		case use_ok:
			must_close= false;
			owner().MR_Send(msg_edited);
			Editor_Window_Apply(); 
			break;

		default:
			break;
		}
		break;

	case msg_reset:
		Editor_Window_New(); 
		break;

	default:
		V_Packer::MR_Receive(msg);
		break;
	}
}

#endif /* ! EDITOR_WINDOW_INCLUDED */
