#ifndef EVALUATOR_INCLUDED
#define EVALUATOR_INCLUDED

#include "input_reader.hh"
#include "int.hh"
#include "io_child_process.hh"
#include "timer.hh"
#include "position.hh"
#include "generator.hh"
#include "evaluator_settings.hh"
#include "pointer.hh"

/* Start evaluator in child process.

   Evaluator has three states:

   - Not started:  ! Is()
   - Running:      Is(),  ! Terminated()
   - Finished:     Is(),  Terminated()
 */

class Evaluator
	:  private IO_Child_Process,
	   private Timer,
	   virtual public Pointed
{
private:

	class EC_Reader
		:  public Input_Reader
	{
	public:
		EC_Reader(Pointer <Evaluator> new_evaluator)
			:  evaluator(new_evaluator)
			{
			}
		
	private:

		Pointer <Evaluator> evaluator;
		
		virtual void IR_Input()
			{
				evaluator().Input();
			}
	};

public:

	Evaluator();

	bool Open(const Position &new_position)
		{
			position= new_position;
			settings.Reset(); 
			has_started= false;
			
			return true;
		}
	void Close()
		{
			position= NULL;
			if (generator.Is())
				generator.Free(); 
		}

	virtual void EC_Terminate(bool /* really */ ,
				  bool /* engine_finished */ )
		{
		}		       
	virtual void EC_Timer(int /* date_milliseconds */,
			      int /* time_milliseconds */)
		/* DATE_MILLISECONDS:  Wall clock time.
		   TIME_MILLISECONDS:  CPU time. 
		*/
		{
		}
	virtual void EC_Run()
		{
		}
	virtual void EC_Pause()
		{
		}
	virtual void EC_Depth(int /* depth */ )
		{
		}
	virtual void EC_Main(int /* move */ )
		{
		}
	virtual void EC_Score(int /* score */ )
		{
		}
	virtual void EC_Knots(intmax_t /* knots */ )
		{
		}

	bool Run();
	void Pause();

	bool Paused() const
		{
			return IO_Child_Process::Paused();
		}
	bool Started() const
		{
			return IO_Child_Process::Is();
		}
	int Get_Best_Move() const
		{
			Assert (has_started); 
			return best_move;
		}
	bool Re_Start();
	void Terminate(bool really= true, bool engine_finished= false);

	bool Get_Infinite() const
		{
			return settings.timing ==
				Evaluator_Settings::ti_infinite; 
		}

	void Set_Infinite(bool new_infinite)
		{
			settings.timing= new_infinite 
				? Evaluator_Settings::ti_infinite
				: Evaluator_Settings::ti_time; 
		}
	
	Date Get_Start_Date() const
		{
			return start_date; 
		}

	const IO_Child_Process &Get_IO_Child() const
		{
			return This; 
		}

	int Get_Knots() const
		{
			Assert (has_started);
			return knots;
		}

	int Get_Depth() const
		{
			Assert (has_started);
			return depth; 
		}
	int Get_Score() const
		{
			Assert (has_started);
			return score; 
		}

	const Generator &Get_Generator() const
		{
			return generator; 
		}

	String Get_Program_Name() const
		{
			return program_name; 
		}
	
	void Input(); 

	void Set_Settings(Evaluator_Settings new_settings)
		{
			settings= new_settings; 
		}
	void Set_Auto_Reply(bool auto_reply)
		{
			settings.auto_reply= auto_reply; 
		}
	void Set_Time(int new_time_s)
		{
			settings.timing= Evaluator_Settings::ti_time;
			settings.time_s= new_time_s;
		}

	const Evaluator_Settings &Get_Settings() const
		{
			return settings; 
		}

	int Get_Date_ms() const
		/* Number of ms the evaluator runs. 
		 */
		{
			Assert (IO_Child_Process::Is()); 
			if (IO_Child_Process::Paused())
				return age; 
			else
				return Date_Diff_ms(start_date,
						    Get_Date()); 
		}
	bool Is() const
		{
//			return IO_Child_Process::Is(); 
			return has_started;
		}
	bool Terminated()
		{
			Assert (Is());
			return ! IO_Child_Process::Is();
		}

private:

	enum {milliseconds= 1000};  /* interval */

	Generator generator;
	int best_move;
	bool has_started;
	/* Whether the best move is known.  The evaluator may be
	   running or finished.  
	*/
	Date start_date; 
	/* when not paused */
	int age;  
	/* in milliseconds, when paused */
	int knots;
	int depth;
	int score;
	Evaluator_Settings settings; 
	Pointer <const Position> position;
	EC_Reader output_reader, error_output_reader;
	String program_name;
	/* Last run. 
	 */
	int last_time_ms; 

	void Timer_Alarm();
	void IOCP_Output(String);
	void IOCP_Error_Output(String);
};

#endif /* ! EVALUATOR_INCLUDED */
