#include "evaluator_widget.hh"

#include "error.hh"
#include "format.hh"
#include "inform.hh"
#include "fen.hh"
#include "text_utils.hh"
#include "parse_move.hh"
#include "time.hh"
#include "format_sec.hh"
#include "trl.hh"
#include "keys.hh"
#include "dbg.hh"
#include "score.hh"
#include "notation.hh"

bool
Evaluator_Widget::Open(Pointer <Widget> parent,
		       int new_msg_play,
		       int new_msg_changed,
		       int new_msg_play_best,
		       const Position &new_position)
{
	if (! Framed <Grid_Packer>::Open
	    (parent, 
	     Framed <Grid_Packer> ::Open_Info
	     (Grid_Packer::Open_Info(2))))
		return false;

	if (! Evaluator::Open(new_position))
	{
		goto close;
	}

  	msg_play=       new_msg_play;
	msg_changed=    new_msg_changed;
	msg_play_best=  new_msg_play_best;

	if (! busy_text.Open(This(), Trl("Status:")))
		goto close;
	if (! busy.Open(This()))
		goto close; 

	{  
	const char *text[count]= {"Status:",
				  "Best move:",
				  "Score:",
				  "Knots:",
				  "Wall clock time:",
				  "CPU time:",
				  "Wall clock speed:",
				  "CPU speed:",
				  "CPU share:"};

	for (int i= 0; i < count; ++i)
	{
	  if (! text_widgets[0][i].Open(This(), 
					text[i],
					dt_text,
					st_background,
					al_begin))
			goto close;

	  if (! text_widgets[1][i].Open(This(),
					i == 0 ? Trl("Ready") : "",
					dt_text,
					st_background,
					al_end))
			goto close;
	}
	}

	return true;

 close:

	Widget::Close();
	return false;
}

void
Evaluator_Widget::MR_Receive(int msg)
{
	Dbg_Function();
	Dbg_Int(msg);

	switch (msg)
	{
	default:
	  Framed <Grid_Packer> ::MR_Receive(msg);
		break;

	case msg_run:

		if (Started())
		{
			if (Paused())
			{
				Pause();
				busy.Start();
			}
		}
		else
		{
			if (! Run())
			{
				String program_name= Get_Program_Name();
				String error= Error::Get();
				Inform(Trl("Error"),
				       Format("Cannot run `%s':  %s.",
					      program_name(),
					      error()));
			}
			else
				busy.Start(); 
		}

		break;

	case msg_pause:
		if (Started()) 
		{
			Pause();
			busy.Set(! Paused());
		}
		break;

	case msg_terminate:
		Terminate();
		busy.Stop();
		break;

	case msg_restart:
		Re_Start();
		busy.Start();
		break;

	case msg_change_auto_reply:
		Set_Auto_Reply(! Get_Settings().auto_reply); 
		break;

	case msg_change_infinite:
	{
		Set_Infinite(! Get_Infinite());
		break;
	}

	case msg_settings:
		settings_editor.Open
			(msg_settings_edited,
			 This,
			 Get_Settings()); 
		break;

	case msg_settings_edited:
	{
		Evaluator_Settings es= settings_editor.Get_Value();
		settings_editor.Have_Read();
	
		Set_Settings(es); 

		break;
	}
	}
}

void
Evaluator_Widget::Wdg_Close()
{
	Dbg_Function();

	if (IO_Child_Process::Is())
		Terminate();
	if (Timer::Is())
		Timer::Stop();
 	Evaluator::Close();
	settings_editor.Close();

	Framed <Grid_Packer> ::Wdg_Close();
}

void
Evaluator_Widget::Clear_Info()
{
	if (text_widgets[1][0].Is())
		text_widgets[1][0].Set_Text("Ready");

	for (int i= 1; i < count; ++i)
		if (text_widgets[1][i].Is())
			text_widgets[1][i].Set_Text("");
}

void Evaluator_Widget::
EC_Timer(int date_ms,
	 int time_ms)
{
	/* Display the date only with second precision.  Add one
	   second if TIME would be greater than DATE. 
	*/

	int date_s= date_ms / 1000;
	if (date_s * 1000 < time_ms)
		++ date_s; 

	if (text_widgets[1][4].Is())
		text_widgets[1][4].Set_Text(Format_Sec(date_s)); 
	if (text_widgets[1][5].Is())
		text_widgets[1][5].Set_Text(Format_Sec(time_ms / 1000, true,
						 time_ms % 1000));
	if (text_widgets[1][8].Is())
		text_widgets[1][8].Set_Text
			(Format("%d,%c %%",
				time_ms * 100 / (date_ms >? 1),
				'0' + time_ms * 1000 / (date_ms >? 1) % 10));
}

void
Evaluator_Widget::Menued_Get(int menu, 
			     Menued::Entries &entries)
{
	switch (menu)
	{
	case menu_main:
		entries.Append_Command(msg_run,
				       Trl("Run"),
				       key_run, 
				       gl_run,
				       ! Evaluator::Is());
		bool running= Evaluator::Is() && ! Evaluator::Terminated();

		entries.Append_Checked(msg_pause,
				       Trl("Pause"),
				       running && Evaluator::Paused(),
				       true,
				       running);

		entries.Append_Command(msg_terminate,
				       Trl("Terminate"),
				       key_terminate,
				       gl_nothing,
				       running);
		entries.Append_Command(msg_restart,
				       Trl("Restart"),
				       key_no,
				       gl_nothing,
				       Evaluator::Is());
		entries.Append_Command
			(msg_play_best,
			 Trl("Play main move"),
			 key_no,
			 gl_nothing,
			 Evaluator::Is() && 
			 *Evaluator::Get_Generator());
		entries.Append_Separator();
		entries.Append_Checked(msg_change_auto_reply,
				       Trl("Auto reply"),
				       Get_Settings().auto_reply,
				       false);
		entries.Append_Separator();
		entries.Append_Checked(Get_Infinite() 
				       ? (int)msg_ignore 
				       : (int)msg_change_infinite,
				       Trl("Infinite"),
				       Get_Infinite(),
				       true);
		entries.Append_Checked(Get_Infinite()
				       ? (int)msg_change_infinite
				       : (int)msg_ignore,
				       Trl("Time"),
				       ! Get_Infinite(),
				       true);
		entries.Append_Separator(); 
		entries.Append_Command(msg_settings,
				       Trl("Engine preferences"));
	}
}

void
Evaluator_Widget::EC_Knots(intmax_t knots)
{
	text_widgets[1][3].Set_Text(Format_Intmax(knots));
	int date_ms= Evaluator::Get_Date_ms(); 
	text_widgets[1][6].Set_Text
		(Format_Intmax(knots * 1000 / (date_ms >? 1)) + " KT/s");
	Process_Usage usage(Get_IO_Child().Get_Process_Usage());
	int time_ms= usage.total + usage.children_total; 
	text_widgets[1][7].Set_Text
		(Format_Intmax(knots * 1000 / (time_ms >? 1)) + " KT/s");
}

void
Evaluator_Widget::EC_Terminate(bool really,
			       bool engine_finished)
{
	if (text_widgets[1][0].Is())
		text_widgets[1][0].Set_Text(really 
				      ? "Ready" 
				      : "Finished");
	if (really)
	{
		Clear_Info();
	}
	else
		if (text_widgets[1][0].Is())
			text_widgets[1][0].Set_Text("Finished");
	if (Get_Settings().auto_reply && engine_finished)
		Parent_Send_Message(msg_play);
	Parent_Send_Message(msg_changed);
	busy.Stop();
}

void Evaluator_Widget::EC_Score(int score)
{
	if (text_widgets[1][2].Is())
		text_widgets[1][2].Set_Text(Format_Score(score)); 
}

void Evaluator_Widget::EC_Run()
{
	text_widgets[1][0].Set_Text("Running");
	Parent_Send_Message(msg_changed);
}

void Evaluator_Widget::EC_Pause()
{
	text_widgets[1][0].Set_Text(Get_IO_Child().Paused()
			      ? "Paused"
			      : "Running");
	Parent_Send_Message(msg_changed);
}

void Evaluator_Widget::EC_Depth(int depth)
{
	if (text_widgets[1][0].Is())
		text_widgets[1][0].Set_Text
			(Format("Depth %d",
				depth));
}

void Evaluator_Widget::EC_Main(int move)
{
	if (text_widgets[1][1].Is())
		text_widgets[1][1].Set_Text
			(san_notation.Format_Move_Name
			 (Get_Generator().Get_Move_Name(move)));
						      
}
