#include "game.hh"

#include "text_utils.hh"
#include "parse_move.hh"
#include "fen.hh"
#include "trl.hh"
#include "stream_file.hh"
#include "deref.hh"
#include "read_fen.hh"

Game::Game()
{
}

Game::Game(const Game &game)
	:  Pointed()
{
	This= game;
}

void
Game::Close()
{
	Assert (Is());
	location_stack= 0;

	variation.Free();
	generator.Free();
}

void
Game::Open()
{
	/* Must free VARIATION_STACK before VARIATION because
	   VARIATION_STACK contains pointers on objects owned by
	   VARIATION. 
	 */
	variation_stack.Set(1);
	variation.Free();
	location_stack.Set(1);
	location_stack[0]= 0;
	variation_stack[0]= & variation; 
	position.Reset();
	generator.New(position);
	game_info.Reset(); 
	initial_position.Free();
}

bool
Game::Open(const Position &new_position) 
{
	if (initial_position.Is())
		initial_position()= new_position;
	else
	{
		if (! initial_position.Alloc(new_position))
			return false;
	}

	location_stack.Set(1);
	location_stack[0]= 0;
	variation_stack.Set(1);
	variation_stack[0]= & variation; 

	variation.Free();
	
	if (generator.Is())  generator.Free(); 
	position= initial_position();

	generator.New(position);

	game_info.Reset(); 

	return true;
}

bool
Game::Open(String fen)
{
	Position position;
	if (! Read_FEN(fen, position))
		return false;
  
	if (! Open(position))
		return false;

	return true;
}

bool
Game::Append_Move(int move_index)
{
	Assert(Is());
	Assert(move_index >= 0 &&
	       move_index < *generator);

	Assert(location_stack[-1] == *variation_stack[-1]());
	
	if (! variation_stack[-1]().Append_Move(generator, move_index))
		return false;

	position << generator[move_index];
	if (! generator.New(position))
	{
		position >> generator[move_index];
		return false; 
	}
	++ location_stack[-1]; 

	return true;
}

bool
Game::Go_To_Start()
{
	Dbg_Function();

	Assert (Is());

	if (*location_stack == 1 && location_stack[0] == 0)
		return false;

	location_stack= 1;
	variation_stack= 1;
	location_stack[0]= 0; 
	Reset_Position();
	generator.New(position);

	return true;
}

void
Game::Go_Backward_In_Current()
{
	Assert(Is());

	/* Take back last move from current variation. 
	 */
	position >> variation_stack[-1]()[location_stack[-1] - 1];
	if (! generator.New(position))
	{
		position << variation_stack[-1]()[location_stack[-1] - 1];
		return;
	}
	-- location_stack[-1];
}

bool
Game::Go_Forward()
{
	Dbg_Function();

	Assert (Is());


	int size= location_stack[-1];
	int var_size= *variation_stack[-1]();

	Dbg_Int(size);
	Dbg_Int(var_size); 

	if (size == var_size)
		return false;

	position << variation_stack[-1]()[location_stack[-1]++];
	generator.New(position);

	return true;
}

bool
Game::Go_To_End()
{
	Assert(Is());

	if (location_stack[-1] == *variation_stack[-1]())
		return false;

	do
		position << variation_stack[-1]()[location_stack[-1]++];
	while (location_stack[-1] < *variation_stack[-1]());

	generator.New(position);

	return true;
}

void
Game::Remove_End()
{
	Assert (Is());

	variation_stack[-1]().Cut(location_stack[-1]);
}

Game::~Game()
{
}

void Game::Append_Variation(int move_index)
{
	Append_Variation();
	Append_Move(move_index); 
}

bool Game::
Append_Variation() 
{
	/* The position does not change ; GENERATOR is not changed. 
	 */

	if (! variation_stack[-1]().Append_Variation(location_stack[-1]))
		return false;
	if (! location_stack.Set(*location_stack + 1))
		return false;
	if (! variation_stack.Set(*location_stack))
	{
		location_stack= -1; 
		return false;
	}

	location_stack[-1]= 0;
	int index= location_stack[-2];
	variation_stack[-1]= 
		variation_stack[-2]()
		.Get_Variations(index)[-1]();

	return true; 
}

void Game::Go_Up()
{
	Assert (*location_stack != 1);
	Assert (location_stack[-1] == 0);

	location_stack= -1;
	variation_stack= -1; 
}

void Game::
operator = (const Game &game)
{
	if (Is())
		Close();
	variation= game.variation;
	location_stack= game.location_stack;
	variation_stack.Set(*game.variation_stack);
	position= game.position;
	initial_position= game.initial_position; 
	if (*location_stack)
	{
		generator.New(position);
		variation_stack[0]= variation;
		for (int i= 1;  i < *location_stack;  ++i)
		{
			int n= -1;
			for (int j= 0;  
			     (j < *game.variation_stack[i - 1]()
			      .Get_Variations(location_stack[i - 1]) &&
			      n < 0);
			     ++j)
				if (game.variation_stack[i] == & game.variation_stack[i - 1]()
				    .Get_Variations(location_stack[i -1])[j]())
					n= j;
			variation_stack[i]= variation_stack[i - 1]()
				.Get_Variations(location_stack[i - 1])[n]();
		}
	}
	game_info= game.game_info;
}

void Game::
Reset_Position()
{
	Dbg_Function();
	Dbg_Int(initial_position.Is());

	if (initial_position.Is())
		position= initial_position();
	else
		position.Reset();
}

void Game::
Enter_Variation(int index)
{
	Assert (index >= 0);
	Assert (Get_Current_Variation().Have_Variations(location_stack[-1]));
	Assert (index < *Get_Current_Variation().Get_Variations(location_stack[-1]));

	if (! variation_stack.Set(*variation_stack + 1))
		return;

	if (! location_stack.Set(*location_stack + 1))
	{
		variation_stack= -1;
		return;
	}

	location_stack[-1]= 0;
	variation_stack[-1]= variation_stack[-2]().Get_Variations(location_stack[-2])[index]();
}

Move_Name Game::
Get_Overall_Move_Name(int i) const
{
	Dbg_Function();

	int overall_pos= 0;
	for (int j= 0;  j < *location_stack;  ++j)
	{
		if (j + 1 == *location_stack || i < overall_pos + location_stack[j])
		{
			int current_pos= i - overall_pos;
			Dbg_Int(j);
			Dbg_Int(current_pos);
			Move_Name result=
				variation_stack[j]().Get_Move_Name(current_pos);
			return result;
		}
		overall_pos += location_stack[j];
	}
	Assert_Unreached();

	return Move_Name(); 
}

void Game::Go_To_Overall_Location(int overall_location)
{
	if (overall_location < Get_Overall_Location())
	{
		do
		{
			if (location_stack[-1] == 0)
				Go_Up();
			Go_Backward_In_Current(); 
		}
		while (overall_location < Get_Overall_Location());
		if (location_stack[-1] == 0 && overall_location != 0)
			Go_Up();
		

#if 0
		do 
			Go_Backward_In_Current();
		while (overall_location < Get_Overall_Location());

		while (*location_stack != 1 && location_stack[-1] == 0)
			Go_Up(); 
#endif /* 0 */ 
	}
	else if (overall_location > Get_Overall_Location())
	{
		int previous= Get_Current_Variation().Get_Previous_Move_Count();
		int current_location= overall_location - previous;
		Assert (current_location >= 0 && 
			current_location <= Get_Current_Variation().Get_Length());
		while (location_stack[-1] < current_location)
			position << variation_stack[-1]()[location_stack[-1]++]; 
		generator.New(position);
	}
}

void Game::Remove_Variation()
{
	if (*location_stack == 1)
		return;

	while (location_stack[-1])
		Go_Backward_In_Current();
	int vindex= -1;
	int count= *variation_stack[-2]().Get_Variations(location_stack[-2]); 
	for (int i= 0;  
	     i < count; 
	     ++i)
	{
		if ((& variation_stack[-2]().Get_Variations(location_stack[-2])[i]()) ==
		    (& variation_stack[-1]()))
		{
			vindex= i;
			i= count;
		}
	}
	Go_Up();
	variation_stack[-1]().Remove_Variation(location_stack[-1],
					     vindex); 
	generator.New(position); 
}

void Game::Remove_Variation(int index)
{	
	variation_stack[-1]().Remove_Variation
		(location_stack[-1], index); 
}

const Move &Game::Get_Overall_Move(int location) const
{
	Dbg_Function();
	Dbg_Int(location);

	int overall_previous= 0;
	for (int j= 0;  j < *location_stack;  ++j)
	{
		Dbg_Int(j);
		if (location < overall_previous + location_stack[j])
			return variation_stack[j]()[location - overall_previous];
		overall_previous += location_stack[j];
	}

	/* Requested location is after the location in current
	   variation. 
	*/
	Dbg("\tafter.");
	return variation_stack[-1]()[location - overall_previous];
}

bool Game::Is_Overall_Variation_Move(int overall_location) const
{
	if (*location_stack == 1)
		return false;
	for (int j= 1;  j < *variation_stack;  ++j)
	{
		if (overall_location == variation_stack[j]()
		    .Get_Previous_Move_Count())
			return true;
	}
	return false;
}

void Game::Commentify(String &comment)
{
	for (int i= 0;  i < *comment;  ++i)
		if (comment[i] == '}')
			comment.Get_Ref()[i]= ' '; 
}

void Game::Set_NAG(int nag)
{
	Assert (location_stack[-1] != 0);
	variation_stack[-1]().Set_NAG(location_stack[-1] - 1, nag); 
}

bool Game::Leave_Variation()
{
	if (*variation_stack == 1)
		return false;

	while (location_stack[-1] != 0)
		Go_Backward_In_Current();

	Go_Up();

	return true; 
}

bool Game::Go_Backward()
{
	if (*location_stack == 1 && location_stack[0] == 0)
		return false;

	Go_Backward_In_Current();

	if (location_stack[-1] == 0)
	{
		if (*location_stack != 1)
			Go_Up();
	}

	Assert (*location_stack == 1 || location_stack[-1] != 0);

	return true;
}

bool Game::Make_Main()
{
	while (*location_stack != 1)
		Make_Parent();
	return true; 
}

bool Game::Make_Parent()
{
	if (*location_stack == 1)
		return false;

	/* Index of current subvariation.
	 */
	int loc= location_stack[-2];
	Assert (variation_stack[-2]().Have_Variations(loc));
	int vcount= *variation_stack[-2]().Get_Variations(loc);
	int vindex= -1;
	for (int i= 0;  i < vcount;  ++i)
	{
		if ((& variation_stack[-2]().Get_Variations(loc)[i]()) ==
		    (& variation_stack[-1]()))
		{
			vindex= i;
			i= vcount;
		}
	}
	Assert (vindex >= 0);
	variation_stack= -1;

	/* Detach old main variation.  The current subvariation is
	   still contained in it.  Existing variation is truncated. 
	 */

	Dynamic <Variation> old_main;
	Variation *newv= Data_Alloc
		(Variation,
		 variation_stack[-1](), 
		 location_stack[-2]);
	if (newv == NULL)
		return false;
	old_main.Attach(newv);

	/* Transfer old subvariation from OLD_MAIN to game. 
	 */
	variation_stack[-1]().Append_Moves(old_main().Get_Variations(0)[vindex]());
	old_main().Remove_Variation(0, vindex); 

	/* Append subvariaitons
	 */
	if (old_main().Have_Variations(0))
	{
		for (int i= 0;  
		     i < *old_main().Get_Variations(0); 
		     i++)
		{
			variation_stack[-1]().Append_Variation
				(location_stack[-2],
				 old_main().Get_Variations(0)[i]);
		}
		variation_stack[-1]().Get_Variations(0).Free(); 
	}
	
	/* Append OLD_MAIN as subvariation to current variation. */ 
	variation_stack[-1]().Append_Variation
		(location_stack[-2],
		 old_main);
	Assert (! old_main.Is());

	location_stack[-2] += location_stack[-1];
	location_stack= -1;
	
	/* Normalize */ 

//	variation.Normalize(); 

	return true; 
}

void Game::Normalize()
{
	variation.Normalize();
}

void Game::Play_Variation(int i)
{
	Enter_Variation(i);
	bool f= Go_Forward();
	Assert (f);
}

