#include "ics.hh"

#include "format.hh"
#include "text_utils.hh"
#include "deref.hh"
#include "ding.hh"
#include "generator.hh"
#include "dbg.hh"
#include "notation.hh"

bool
ICS::Open(String host_name,
	  int port,
	  String u,
	  String p)
{
	Dbg_Function();

	Assert (! Is());

	if (! Socket_Child_Process::Open(host_name, port))
	{
		Dbg("\tfail.");
		return false;
	}
	Dbg("\tsuccess.");

	host= host_name;
	user= u;
	passwd= p;
	seeking= false;
	ready= false;
	inside_game= false;
	error= "";

	return true;
}

void 
ICS::IOCP_Output(String line)
{
	Dbg_Function();
	Dbg_Str(line());

	ICS_Message(line);

	if (! line.Is())
		return;

	const char *p= line();

	p += Skip_Spaces(p);

	if (Compare_Words(p, "login:"))
	{
		Write(user);
		Write(passwd, false);
		ICS_Message("<Sent password>");
		Write("");
		Write("style 12"); /* Set to style #12, as does
				      XBoard. */

		ready= true;
		if (seeking)
			Write_Seek();
	}
	else if (Compare_Words(p, "<12>"))
	{
		p += Get_Next_Word(p);
		Style_12(p);
	}
	else if (p[0] == '{')
	{
	  inside_game= false;

		while (p[0] != '}' && p[0] != '\0')
			++ p;
		if (p == '\0')
			return;
		++p;
		p += Skip_Spaces(p);
		if (Compare_Words(p, "1-0"))
			result= 2;
		else if (Compare_Words(p, "0-1"))
			result= 0;
		else if (Compare_Words(p, "1/2-1/2"))
			result= 1;
		else
			result= 3;

		ICS_Finished();
	}
}

void 
ICS::IOCP_Error_Output(String s)
{
	Dbg_Function();
	Dbg_Str(s());

	ICS_Message(s);

	if (! error())
		error= s;
}

void
ICS::Seek(int m, int s)
{
	Assert (Is());

	seeking= true;
	seek_m= m;
	seek_s= s;

	if (ready)
		Write_Seek();
}

void
ICS::Write_Seek()
{
	Assert (ready);

	Write(Format("seek %d %d",
		     seek_m,
		     seek_s));
}

int
ICS::Style_12(const char *p)
{
	Assert (Is());

	Position pos;

	pos.Get_Material().Clear();

	for (int rank= 7; rank >= 0; -- rank)
	{
		for (int file= 0; file < 8; ++ file)
		{
			if (p[file] == '\0')
				return 1;
			if (p[file] == '-')
			  pos.Get_Piece(Make_Square(file,rank))= 
			    Piece(false);
			else
			{
				int new_pt= San_Char(p[file]);
				if (new_pt == 8)
					return 2;
				int new_cc= p[file] >= 'a' && p[file] <= 'z' ? cc_b : cc_w; 
				pos.Get_Piece(Make_Square(file, rank))= 
				  Piece(new_cc, new_pt);

				++ pos.Get_Material()(new_pt, new_cc);
				if (new_pt == pt_K)
				  {
				    if (pos.Get_Material(pt_K, new_cc) > 1)
				      /* more than one king on one side */ 
				      return 3;
				    pos.Get_King_Square(new_cc)= Make_Square(file, rank);
				  }
			}
		}
		p += Get_Next_Word(p);
	}

	if (! (pos.Get_Material(pt_K, cc_w) && pos.Get_Material(pt_K, cc_b)))
		return 5;

	if (p[0] == 'B')
		pos.Set_To_Move(cc_b);
	else if (p[0] == 'W')
		pos.Set_To_Move(cc_w);
	else
		return 6;

	p += Get_Next_Word(p);

	if (Compare_Words(p, "-1"))
		pos.Get_EP_Square()= 0;
	else if (p[0] >= '0' &&
		 p[0] <= '7')
		pos.Get_EP_Square()= Make_Square
			(p[0] - '0', 
			 pos.Get_To_Move() == cc_b ? sq_4 : sq_5);
	else 
		return 7;

	bool castle[4];
	for (int i= 0; i < 4; ++i)
	{
		p += Get_Next_Word(p);
		if (p[0] == '0')
			castle[i]= false;
		else if (p[0] == '1')
			castle[i]= true;
		else 
			return 8;
	}

	pos.Get_Moved(0)= castle[1] ? 0 : 1;
	pos.Get_Moved(1)= castle[0] || castle[1] ? 0 : 1;
	pos.Get_Moved(2)= castle[0] ? 0 : 1;
	pos.Get_Moved(3)= castle[3] ? 0 : 1;
	pos.Get_Moved(4)= castle[2] || castle[3] ? 0 : 1;
	pos.Get_Moved(5)= castle[2] ? 0 : 1;

	p += Get_Next_Word(p);
	p += Get_Next_Word(p);

	p += Get_Next_Word(p);
	if (p[0] == '\0')
		return 9;

	String white(p, Skip_Word(p));
	p += Get_Next_Word(p);
	if (p[0] == '\0')
		return 10;
	String black(p, Skip_Word(p));
	
	p += Get_Next_Word(p);

	if (p[0] == '\0')
		return 11;

	Chess_Color me;
	bool new_observing= false;
	if (Compare_Words(p, "1"))
		me= pos.Get_To_Move();
	else if (Compare_Words(p, "-1"))
		me= CC_Turn(pos.Get_To_Move());
	else
		new_observing= true;

	for (int i= 0; i < 5; ++i)
	{
		p += Get_Next_Word(p);
		if (p[0] == '\0')
			return 12;
	}

	String white_clock(p, Skip_Word(p));
	p += Get_Next_Word(p);
	if (p[0] == '\0')
		return 13;
	String black_clock(p, Skip_Word(p));
	p += Get_Next_Word(p);
	if (p[0] == '\0')
		return 14;

	int white_time, black_time;
	if (1 != sscanf(white_clock(), "%d", & white_time))
		return 15;
	if (1 != sscanf(black_clock(), "%d", & black_time))
		return 16;

	for (int i= 0; i < 3; ++i)
	{
		p += Get_Next_Word(p);
		if (p[0] == '\0')
			return 17;
	}

	String move= String(p, Skip_Word(p));

	if (new_observing || pos.Get_To_Move() == me)
		Ding();

	/* allow non-normalized positions */

	if (inside_game)
	{
		if (white_player != white ||
		    black_player != black)
		  inside_game= false;
		else if (new_observing || pos.Get_To_Move() == play_color)
			ICS_Playing(move);
	}
	position= pos;
	generator.New(position); 

	if (! inside_game)
		/* begin  new game */
	{
	  inside_game= true;
		observing= new_observing;
		white_player= white;
		black_player= black;
		play_color= me;
		ICS_New_Game();
	      }

	ICS_Changed();
	ICS_Clock(white_time, black_time);
	return 0;

}

bool
ICS::User_Play(int from, 
	       int to)
{
	Assert (Is());

	if (inside_game)
	{
		if (play_color != position.Get_To_Move())
			return false;
		Generator generator;
		generator.New(position);
		int sm= generator.Get_Square_Move(from, to);
		if (sm == -1)
		{
			return false;
		}
		else
		{
			Write(Format_Square(from) + Format_Square(to));
			position << generator.Get_Move(sm);
			generator.New(position); 
			ICS_Changed();
			return true;
		}
	}
	else
		return false;
}

bool
ICS::Proceed() 
{
	return Socket_Child_Process::Proceed(); 
}

void
ICS::Write(String s,
	   bool _echo)
{
	Socket_Child_Process::Write(s);
	if (_echo)
		ICS_Message(s);
}

