#ifndef NOTATION_INCLUDED
#define NOTATION_INCLUDED

/* Chess notation. 
 */

#include "string.hh"
#include "move_name.hh"

/* Notation options. 
 */
enum
{nt_short             = (1 << 0), 
 /* Short notation */ 
 nt_pawn_sign         = (1 << 1),
 /* Sign for pawn moves */
 nt_ep                = (1 << 2),
 /* En passant */
 nt_check             = (1 << 3),
 /* Check, checkmate */
 nt_short_capture     = (1 << 4),
 /* Capture in short notation */ 
 nt_short_capture_end = (1 << 5),
 /* Capture sign at the end in short notation */
 nt_short_pawn        = (1 << 6),
 /* Sign for pawn captures in short notation */ 
 nt_capture_name      = (1 << 7),
 /* Name of captured piece in long notation */ 
};

class Notation
	:  public Pointed
{
public:

	Notation() 
		/* Raw */ 
		{
		}
	Notation(char new_pieces[6],
		 char new_capture,
		 char new_promotion,
		 int new_options)
		:  capture(new_capture),
		   promotion(new_promotion),
		   options(new_options)
		{
			for (int i= 0;  i < 6;  ++i)
				pieces[i]= new_pieces[i]; 
		}

	int Get_Options() const
		{
			return options; 
		}
	int &Get_Options()
		{
			return options; 
		}
	void Set_Options(int new_options)
		{
			options= new_options; 
		}

	String Format_Move_Name(Move_Name) const;

	char Format_Piece_Type(Piece_Type piece_type) const
		{
			Assert (piece_type >= 0 && piece_type < 6);
			return pieces[piece_type];
		}


	void Reset();
	char *Get_Pieces()
		{
			return pieces;
		}
	char &Get_Capture()
		{
			return capture;
		}
	char &Get_Promotion()
		{
			return promotion; 
		}
	const char *Get_Pieces() const
		{
			return pieces; 
		}
	static const char *Get_RC_Name()
		{
			return "notation"; 
		}

private: 

	char pieces[6];
	/* Characters for each piece type. 
	 */
	char capture;
	/* Sign for capture or '\0'. 
	 */
	char promotion;
	/* Sign for promotion or '\0'. 
	 */

	int options; 
	/* nt_* constants. 
	 */

	void Append_Sign(String &string, 
		      Piece_Type piece_type) 
		const
		{
			Assert (piece_type >= 0 && piece_type < 6); 
			if ((piece_type != pt_P) || (options & nt_pawn_sign))
				string += pieces[piece_type];
		}
};

extern
const Piece_Type
san_chars[26];
/* For each letter from A to Z, the piece type that corresponds to the
   letter.  If the letter 'x' is a character for a piece in SAN, then
   san_chars['x'] contains the piece type.  If the characters is not
   used in SAN, then the character might correspond to a piece type in
   another language.  Letters that do not match any languages are
   filled with 8. 
 */

Inline
Piece_Type
San_Char(char c)
	/* Interface to SAN_CHARS. 
	 */
{
	if (c >= 'A' && c <= 'Z')
		c += ('a' - 'A');

	if (! (c >= 'a' && c <= 'z'))
		return 8;

	return san_chars[c - 'a'];
}

extern
Notation
san_notation;
/* The standard algebraic notation, as defined by the PGN standard. 
 */

#endif /* ! NOTATION_INCLUDED */
