#ifndef OWNED_EDITOR_INCLUDED
#define OWNED_EDITOR_INCLUDED

#include "dynamic.hh"
#include "widget.hh"

/* TYPE is the editor window type.  TYPE should be derived from
   Editor_Window <T>.  TYPE should define Get_Value() as a way to
   access the data in write protected mode.  Type::Edited must be
   defined as the type that is edited.   
 */

template <class Type>
class Owned_Editor 
	:  public Dynamic <Type>
{
public:

	~Owned_Editor()
		{
			Assert (! Is());
		}

	void Open(int msg,
		  Widget &owner)
		{
			typename Type::Edited edited;
			edited.Reset();
			Open(msg, owner, edited);
		}

	void Open(int msg,
		  Widget &owner, 
		  const typename Type::Edited &edited,
		  Pointer <Widget> parent_window= NULL);

	const typename Type::Edited &Get_Value() 
		{
			return (*this)().Editor_Window_Get_Value();
		}

	void Close();
	/* Always force */

	bool Have_Read();
	/* Called after the command is received.  Try to close
	   the window.  Return TRUE if window was closed.
	   Return FALSE if the window is still open.  
	*/

	bool Is_Editing() const
		{
			return Is() && This().Is(); 
		}
};

template <class Type>
void Owned_Editor <Type> ::
Open(int msg,
     Widget &owner, 
     const typename Type::Edited &edited,
     Pointer <Widget> parent_window)
{
	if (Dynamic <Type> ::Is())
	{
		if (This().Is())
		{
			This().Activate_Window();
			return;
		}
		else
		{
			Dynamic <Type> ::Free();
			Dynamic <Type> ::Alloc();
		}
	}
	else
	{
		Dynamic <Type> ::Alloc();
	}
	if (Dynamic <Type> ::Is())
	{
		This().Open(NULL, msg, owner, edited, parent_window);
	}
}

template <class Type>
bool Owned_Editor <Type> ::Have_Read()
	/* Called after the command is received.  Try to close
		   the window.  Return TRUE if window was closed.
		   Return FALSE if the window is still open.  
		 */
{
	if (! Is())
		return true;
	if (! This().Is())
	{
		Free();
		return true; 
	}
	if (This().Close_Editor(false))
	{
		Free(); 
		return true;
	}
	else
	{
		return false; 
	}
}

template <class Type>
void Owned_Editor <Type> ::Close()
{
	if (Is())
	{
		if (This().Is())
		{
			bool closed= This().Close_Editor(true);
			Assert (closed);
		}
		Free();
	}
}

#endif /* ! OWNED_EDITOR_INCLUDED */
