#ifndef PIECE_INCLUDED
#define PIECE_INCLUDED

/* Chess piece.
 */

#include "chess_color.hh"
#include "piece_type.hh"
#include "int.hh"

class Piece
/* A chess piece of a specific color, or an "empty square", or a
   square outside the chess board. 
 */
{
public:

	Piece() 
		/* Raw 
		 */
		{
		}

	Piece(Chess_Color chess_color, 
	      Piece_Type piece_type)
		/* Not empty and not outside. 
		 */
		{
			Assert (chess_color == cc_w  || chess_color == cc_b);
			Assert (piece_type >= 0 && piece_type < 6);

			value= (8 << chess_color) | piece_type; 
		}

	Piece(bool outside)
		/* OUTSIDE=
		   FALSE:  empty square
		   TRUE :  outside the chess board
		*/
		{
			if (outside)
				value= 64;
			else /* empty */
				value= 32;
		}

	bool operator == (Piece piece) const
		{
			return value == piece.value;
		}

	bool Is() const 
		/* Whether THIS is a piece.  FALSE when an empty square or a
		   square outside the chess board. 
		*/
		{
			return value & 24;
		}

	Piece_Type Get_Type() const
		/* Must not be empty or outside.
		 */
		{
			Assert (Is());
			return value & 7;
		}

	bool Is_Type(Piece_Type piece_type) const 
		{
			return Is() && Get_Type() == piece_type;
		}

	bool Is_Empty() const 
		/* Whether THIS is an empty piece. */
		{
			return value == 32;
		}

	Chess_Color Get_Color() const
		/* must be an actual piece */
		{
			Assert(Is());
			return (Chess_Color)(value >> 4);
//			return (Chess_Color)(value >> 3);
		}

	bool Is_Piece_Of_Color(Chess_Color color) const
		{
			Assert (color == cc_w || color == cc_b);
			return value & (8 << color);
		}
	bool Is_Piece_Of_Type(Piece_Type piece_type) const
		{
			Assert (piece_type >= 0 && piece_type < 6);
			return (value & ~24) == piece_type; 
		}

	bool Can_Color_Move_To(Chess_Color color)
		/* Whether a piece of COLOR can move to square with
		   this piece;  empty squares and out-of-board squares
		   are recognized. 
		*/
		{
			Assert (color == cc_w || color == cc_b);
			return value & (32 | (8 << (color ^ 1))); 
		}

	bool operator != (Piece piece) const
		{
			return value != piece.value;
		}

private:
	
	uint8_t value;
	/* Values:

	   8...13:   white pieces
	   16...21:  black pieces
	   32:       empty square
	   64:       outside

	   Bits:

	   0...2:    piece type
	   3:        white
	   4:        black
	   5:        empty
	   6:        outside
	*/
};

static const Piece pi_wP(cc_w, pt_P);
static const Piece pi_wR(cc_w, pt_R);
static const Piece pi_wK(cc_w, pt_K);
static const Piece pi_bP(cc_b, pt_P); 
static const Piece pi_bR(cc_b, pt_R);
static const Piece pi_bK(cc_b, pt_K);
static const Piece pi_empty(false); 
static const Piece pi_outside(true); 

#endif /* ! PIECE_INCLUDED */
