#include "process.hh"

//#ifdef PROCESS_DEFAULT

#include "assert.hh"
#include "deref.hh"
//#include "parse.hh"
#include "text_utils.hh"
#include "error.hh"
#include "io.hh"
#include "format.hh"
#include "debug.hh"
#include "file_name.hh"

#ifndef NDEBUG
#    include "env.hh"
#endif

#include <unistd.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>
#include <limits.h>

#ifndef PRIO_MAX
#    ifdef NZERO
#        define PRIO_MAX ((2 * NZERO) - 1)
#    else
#        define PRIO_MAX 19
#    endif
#endif

void
Process_FD_Set(int fd)
	/* Added as hook to File. 
	 */
{
	/* Set close on exec flag on each FD opened. 
	 */
	fcntl(fd, F_SETFD, FD_CLOEXEC);
}

static
class Process_Main
{
public:

	Process_Main()
	{
		Stream::Set_FD_Creation_Hook(& Process_FD_Set);
	}
} process_main;

extern char **environ;

bool
Process_Create(String program_file_name, 
	       const Vector <String> &args_vector,
	       bool wait,
	       Process *process,
	       Pointer <Stream> input,
	       Pointer <Stream> output,
	       Pointer <Stream> error_output,
	       bool nice)
{
	Assert (wait || process != NULL);
	Assert (program_file_name.Is());

	Dbg("Process_Create()");
	Dbg_Str(program_file_name());

#ifndef NDEBUG	
	Dbg("\targs=");
	for (int i= 0;  i < *args_vector;  ++i)
		Dbg("\t args[%d]=  \"%s\"", i, args_vector[i]()); 
	Dbg("\targs end.");
#endif /* ! NDEBUG */ 

	/* Set info for debug output for the child.  
	 */
#ifndef NDEBUG
	static bool env_set= false;
	if (! env_set)
	{
		env_set= true;
		Env::Set(IO_ENV, Format("%d", IO::dbg.Get_FD()));
	}
#endif

	/* First check if PROGRAM is executable.  If it is not, only the child
	   process would know, and this function would return TRUE.
	   Process_Create() should return false if the program doesn't exist.
	   Relative file names cannot be checked.  
	*/

	if (program_file_name[0] == '/')
	{
		Dbg("\tChecking whether `%s' is executable...",
		    program_file_name());
		if (0 == access(program_file_name(), X_OK))
			Dbg("\t\tyes.");
		else
		{
			Dbg("\t\tno.");
			return false;
		}
	}
	else
	{
		Dbg("\tAssuming that `%s' is executable.",
		    program_file_name());
	}


	pid_t pid= fork();

	switch (pid)
	{
	case -1: /* error */
		Dbg("\tfork() failed.");
		return false;

	case 0: /* child */
	{
		Dbg("\t[child] Entering child.");

		/* Pause in child.
		 */
#ifndef NDEBUG
		if (Env::Get("BABY_PROCESS_PAUSE").Is())
		{
			pause();
		}
#endif /* ! NDEBUG */

		/* Set nice.
		 */
		if (nice)
			::nice(PRIO_MAX);

		if (input != NULL)
			dup2(input().Get_FD(),        0);
		if (output != NULL)
			dup2(output().Get_FD(),       1);
		if (error_output != NULL)
			dup2(error_output().Get_FD(), 2);

		/* Build argument list. 
		 */
		const char *args[*args_vector + 2];
		for (int i= 0;  i < *args_vector;  ++i)
			args[i + 1]= args_vector[i]();
		args[*args_vector + 1]= NULL;

		String base_name= Base_Name(program_file_name);
		args[0]= base_name(); 
		Dbg("\t[child] Trying to execute `%s'.",
		    program_file_name());
		execvp(program_file_name(),
		       (char *const *)args);
		Dbg("\t[child] failed.");

		/* execvp() only returns on error.
		 */

		Write_Formatted(IO::err,
			    "\t[child] *** error:  cannot execute `%s':  %s.\n",
			    program_file_name(),
			    Error::Get()());

		_exit(1);
		/* Must use `_exit()' instead of `exit()', else
		   destructors would be called in the forked process.  
		 */
		Assert_Unreached();
		return false;
	}

	default: /* parent */

		if (wait)
			waitpid(pid, NULL, 0);
		else
		{
			Deref(process)= pid;

//#ifndef NDEBUG
			Dbg("Started PID= %d.", (int)pid); 
//#endif /* ! NDEBUG */

#ifdef PROCESS_CREATE_WAIT
			Dbg("\tWaiting.");
			sleep(5);
#endif			
		}
		Dbg("\tsuccess.");
		return true;
	}
}

void
Process_Kill(Process process)
{
	Dbg_Function();
	Dbg_Int((int)process);

	kill(process, SIGKILL);

	waitpid(process, NULL, 0);
}

bool
Process_Wait(Process process, 
	     int &exit_code)
{
	Assert (process != PROCESS_NO);

	return -1 != waitpid(process, & exit_code, 0);
}

bool
Process_Scan(Process &process,
	     String name)
{
	int pi= process;
	return (1 == sscanf(name(),
			  "%d",
			   & pi));
}

bool
Process_Exists(Process process)
{
	return kill(process, SIGUSR1) == 0;
}

Process
Process_Self()
{
	return getpid();
}

String 
Process_Format(Process process)
{
	return Format("%d", (int) process);
}

void
Process_Pause(Process process)
{
	kill(process, SIGSTOP);
}

void
Process_Resume(Process process)
{
	kill(process, SIGCONT);
}

//#endif /* PROCESS_DEFAULT*/
