#include "read_fen.hh"

#include "parser.hh"
#include "error.hh"
#include "trl.hh"
#include "format.hh"
#include "dbg.hh"
#include "notation.hh"
#include "printify_char.hh"

bool
Read_FEN(String    text,
	 Position &position)
{
	Dbg_Function();

	Assert (text.Is()); 

	Parser parser(text);

	/* first field */
	parser.Skip_Space();
	String first_field= parser.Skip_Word();
	
	position.Get_King_Square(cc_w)= 0;
	position.Get_King_Square(cc_b)= 0;

	int file= 0, rank= 7;
	for (int i= 0;  i < *first_field;  ++i)
	{
		char c= first_field[i];
		switch (c)
		{
		case '0' ... '8':
		{
			int n= c - '0';
			if (file + n > 8)
			{
				Error::Set(Trl("More than eight squares on one row"));
				return false;
			}
			for (int j= 0;  j < n;  ++j)
				position.Get_Piece(Make_Square(file + j, rank))= Piece(false);
			file += n;
			break;
		}
		case 'a' ... 'z':
         	case 'A' ... 'Z':
		{
			if (file > 7)
			{
				Error::Set(Trl("More than eight squares on one row"));
				return false;
			}
			Chess_Color color= c >= 'A' && c <= 'Z' ? cc_w : cc_b;
			Piece_Type piece_type= San_Char(c);
			if (piece_type == 8)
			{
			  String printified= Printify_Char(c);
				Error::Set(Format(Trl("Invalid character `%s' in the first field"),
						  printified()));
				return false;
			}
			if (piece_type == pt_P && (rank == 0 || rank == 7))
			{
				Error::Set(Trl("Pawn on the first or last row"));
				return false;
			}
			Piece piece= Piece(color, piece_type);
			int square= Make_Square(file, rank);
			position.Get_Piece(square)= piece;
			if (piece_type == pt_K)
			{
				if (position.Get_King_Square(color))
				{
					Error::Set(Trl("More than one king on one side"));
					return false;
				}
				position.Get_King_Square(color)= square;
			}
			++ file;
			break;
		}

		case '/':
		{
			if (file < 7)
			{
				Error::Set(Trl("Less than eight squares on one row"));
				return false;
			}
			if (rank == 0)
			{
				Error::Set(Trl("More than eight rows"));
				return false;
			}
			file= 0;
			-- rank;
			break;
		}
		
	        default:
		{
			String printified= Printify_Char(c);
			Error::Set(Format(Trl("Invalid character `%s' in the first field"),
					  printified()));
			return false;
		}
	        }
	}

	Assert(file <= 8);
	if (rank > 0)
	{
		Error::Set(Trl("Less than eight rows"));
		return false;
	}

	if (file < 8)
	{
		Error::Set(Trl("Less than eight squares on one row"));
		return false;
	}

	if (! (position.Get_King_Square(cc_w) != 0     &&
	       position.Get_King_Square(cc_b) != 0))
	{
		Error::Set(Trl("No king on one side"));
		return false;
	}

	/* second field */
	parser.Skip_Space();
	String second_field= parser.Skip_Word();
	position.Set_To_Move(second_field[0] == 'w' ? cc_w : cc_b); 

	/* third field */
	parser.Skip_Space();
	String third_field= parser.Skip_Word();
	for (int i= 0; i < 6; ++i)
		position.Get_Moved(i)= 1;
	for (int i= 0;  i < *third_field;  ++i)
	{
		char c= third_field[i];
		if (c == 'K'                             && 
		    position.Get_Piece(sq_e1) == pi_wK   &&
		    position.Get_Piece(sq_h1) == pi_wR)
		{
			position.Get_Moved(1)= 0;
			position.Get_Moved(2)= 0;
		}
		else if (c == 'Q'                           &&
			 position.Get_Piece(sq_a1) == pi_wR &&
			 position.Get_Piece(sq_e1) == pi_wK)
		{
			position.Get_Moved(1)= 0;
			position.Get_Moved(0)= 0;
		}
		else if (c == 'k'                           &&
			 position.Get_Piece(sq_e8) == pi_bK &&
			 position.Get_Piece(sq_h8) == pi_bR)
		{
			position.Get_Moved(4)= 0;
			position.Get_Moved(5)= 0;
		}
		else if (c == 'q'            &&
			 position.Get_Piece(sq_a8) == pi_wR  &&
			 position.Get_Piece(sq_e8) == pi_wK)
		{
			position.Get_Moved(4)= 0;
			position.Get_Moved(3)= 0;
		}
	}

	/* Fourth field:  ep capture square (eg "e6")
	 */
	parser.Skip_Space();
	String fourth_field= parser.Skip_Word();
	Dbg_Str(fourth_field());
	char c= fourth_field[0];
	position.Get_EP_Square()= ((c >= 'a' && c <= 'h') 
				   ? Make_Square(c - 'a', position.Get_To_Move() == cc_w ? 4 : 3) 
			     : 0);
	Dbg_Int(position.Get_EP_Square());

	/* Fifth field:  halfmove clock.  (ignored)
	 */
	parser.Skip_Space();
	String hm_text= parser.Skip_Word();
	int new_hm;
	if (1 != sscanf(hm_text(),
		      "%d",
		      & new_hm))
		new_hm= 0;
	else
		if (new_hm < 0)
			new_hm= 0;
	position.Set_Halfmove_Clock(new_hm); 

	/* Sixth field:  fullmove number
	 */
	parser.Skip_Space();
	String sixth_field= parser.Skip_Word();
	int fullmove;
	if (1 != sscanf(sixth_field(),
		      "%d",
		      & fullmove))
		fullmove= 1;

	position.Set_Previous_Move_Count( (fullmove - 1) * 2 + position.Get_To_Move());

	/* Rest
	 */

	Chess_Color color= CC_Turn(position.Get_To_Move());
	int square= position.Get_King_Square(color);
	if (position.Square_Attacked(square, color))
	{
		String square_text= Format_Square(square);
		Error::Set(Format(Trl("King on %s is in capture"),
				  square_text()));
		return false;
	}

	position.Update_Material();

	return true;
}

