#ifndef STREAM_INCLUDED
#define STREAM_INCLUDED

/* See `doc/stream.hh' for documentation. 
 */

#include "string.hh"
#include "date.hh"
#include "moveable.hh"

#include "stream.config.hh"

#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#ifdef STREAM_HAVE_UNLOCKED
#    define STREAM_FUNCTION(name)  name ## _unlocked
#else
#    define STREAM_FUNCTION(name)  name
#endif

class Stream
	:  public Pointed
{
public:

	enum
	{
		text= 0, binary, type_count,
		read= 0, write, append, mode_count,
	};

	enum {p_read=    1 << 0,
	      p_write=   1 << 1,
	      p_execute= 1 << 2,
	      p_extra=   1 << 3};

	Stream()
		: type(-1), 
		file(NULL) 
		{
		}
	Stream(FILE *f)
		/* F is one of the std... variables.
		 */
		: type(0), file(f) 
		{
			setvbuf(f,
				NULL,
				_IOLBF,
				0x80);
		}
	Stream(int fd);
	~Stream();

	bool Is() const 
		{
			return file != NULL; 
		}
	bool Open(String, int, int= binary);
	void Open_FD(int fd);
	bool Close()
		{
			Assert(type == -1 && file != NULL);

			int fc= fclose(file);
			file= NULL;
			return 0 == fc;
		}

	int Read(void *data, int length)
		{
			Assert (Is());
			return STREAM_FUNCTION(fread)(data, length, 1, file);
		}

	bool Write(const void *data, int length)
		{
			Assert (Is());
			return STREAM_FUNCTION(fwrite)(data, length, 1, file);
		}
	int Read_Line(char *data, int length)
		{
			Assert (Is());
			return STREAM_FUNCTION(fgets)(data, length, file) != NULL;
		}
	bool Write_Line(const char *s)
		{
			Assert (Is());
			Assert (s != NULL);
			return EOF != STREAM_FUNCTION(fputs)(s, file);
		}

	char Read_Char()
		{
			Assert(Is());
//#ifdef STREAM_HAVE_UNLOCKED
			return STREAM_FUNCTION(getc)(file);
//#else
//			return getc(file);
//#endif
		}

	bool Write_Char(char c)
		{
			Assert(Is());
//#ifdef STREAM_HAVE_UNLOCKED
			return EOF != STREAM_FUNCTION(putc)(c, file);
//#else
//			return EOF != putc(c, file);
//#endif
		}

	void Flush()
		{
			Assert(Is());
			fflush(file);
		}


	void Set_Block(bool);

	void Set_FILE(FILE *f)
		{
			file= f;
		}
	static bool Remove(String);
	static bool Access(String file_name, int t)
		{
			Assert (file_name.Is()); 
			int amode;
			switch (t)
			{
			case p_read:       amode= R_OK; break;
			case p_write:      amode= W_OK; break;
			case p_execute:    amode= X_OK; break;
			default:  Assert_Unreached();   return false;
			}
			return 0 == access(file_name(), amode);
		}

	/* extensions */

	FILE *Get_FILE() const
		/* Stream must be open.
		 */
		{
			Assert (file != NULL);
			return file; 
		}

	int Get_FD() const
		/* Must be */
		{
			Assert (Is() || type >= 1);
			return (type >= 1
				? type - 1
				: fileno(Get_FILE()));
		}

	bool Dup(int fd, const char *mod);
	/* Stream must be closed.  Duplicate FD.  Return TRUE on success, FALSE on
	   failure.  MOD is the mode passed tp fdopen(). 
	*/

	bool Open(int _fd,
		  const char *_mod);

	bool Has_FD() const
		/* Whether there is a file descriptor.
		 */
		{
			return type != -1 || file != NULL;
		}

	bool End_Of_File() const
		{
			Assert (Is());
		
			return 0 != feof(file);
		}
	bool Write_String(String text)
		{
			return Write(text(), *text); 
		}
	
	static void Set_FD_Creation_Hook(void (*)(int fd));
	/* Function that will be called on each FD created. 
	 */

private:

	int type;
	/*  0:     standard input/output, and FILE contains STDIN...
	    -1:    owned file.
	    >= 1:  1+FD, where FD is not owned.
	*/
	FILE *file;
	/* NULL when closed or TYPE >= 1.
	 */

	bool Do_Open(String file_name,
		     int mode,
		     int type); 
	/* Open FILE_NAME.  If successfull, set FILE and return
	   TRUE, else return FALSE.   
	*/

};

void
Stream_Call_Hook(int fd);
/* Other modules that open files call this functions.  With the FD
   created as argument. 
 */

#endif /* ! STREAM_INCLUDED */
