#ifndef STRING_INCLUDED
#define STRING_INCLUDED

#include "mem.hh"
#include "pointer.hh"
#include "this.hh"
#include "constant.hh"

#include <string.h>

class String
	:  public Pointed
{
public:

	class Head
	{
	public:
			
		int length;
		/* Length of string whithout '\0'.  The string
		   may contain embedded '\0' characters.
		*/
		int dummy;
		/* So that sizeof(Head) aligns. 
		 */
		int ref_count;
		/* Reference count. 
		 */
		int place;
		/* Maximum size of string that fits into this
		   whithout resizing. 
		*/
		char string[0];
	};

	String()
		:  raw(NULL) 
		{
		}
	String(const String &);
	String(const char *);
	String(const char *, 
	       int size);
	/* If the size is known. 
	 */
	String(char);
	String(int size, int dummy);
	/* New string of size SIZE, not initialized.  The resulting string
	   may contain '\0' characters. 
	*/
	String(const char *a, const char *b)
		{
			raw= NULL;
			This= a;
			This += b; 
		}
	String(const char *a, const String &b)
		{
			raw= NULL;
			This= a;
			This += b;
		}

	String(void *new_raw)
		:  raw(new_raw)
		{
			if (raw != NULL && ! Constant(raw))
			{
				Assert (Deref(head).ref_count >= 1); 
				++ Deref(head).ref_count; 
			}
		}

	~String() 
		{
			Free(raw); 
		}

	bool Is() const 
		{
			return raw != NULL; 
		}

	const char *Get_Str() const
		/* The string must not be empty. 
		 */
		{
			Assert(Is()); 
			return head->string; 
		}

	static const char *Get_String(const void *raw)
		{
			return (raw == NULL     ? ""                 :
				Constant(raw)   ? (const char *)raw  :
				Deref((const Head *)raw).string);
		}
	const char *operator () () const 
		/* Return "" if the string is empty.  Never return
		   NULL.  
		*/
		{
			return Get_String(raw);
		}

	char *Get_Ref()
		/* String must be full and isolated.  Can be written
		   to until the String is changed or destroyed.  Use
		   with care. 
		*/
		{
			Assert (Is());
			Assert (Is_Isolated());
			return Deref(head).string;
		}

	int Get_Length() const 
		{
			return (raw == NULL   ? 0 :
				Constant(raw) ? strlen(constant) :
				Deref(head).length);
		}
	int operator * () const
		{
			return Get_Length();
		}
	char operator [] (int n) const
		/* Get the Nth character of the string.  N can range
		   from -LENGTH to LENGTH.  When N is LENGTH, '\0' is
		   returned.  This[0] is always defined.  For N < 0,
		   characters are counted from the end.  This[-LENGTH]
		   = This[0]. 
		*/
		{
			Assert(n >= - Get_Length() && n <= Get_Length());

			if (raw == NULL)
				return '\0';
			else 
			{
				if (n < 0)
					n += Get_Length();
				if (Constant(raw))
					return constant[n];
				else
					return Deref(head).string[n];
			}
		}

	const char *operator () (int n) const
		{
			Assert (n >= 0 && n <= Get_Length());
			if (raw == NULL)
				return "";
			else if (Constant(raw))
				return constant + n;
			else 
				return Deref(head).string + n;
		}

	void New(int);
	/* New length and file with spaces.  Lose data. 
	 */
	void New_New(int);
	/* new size but whithout initialisation */
	void Free()
		{
			Free(raw);
			raw= NULL;
		}
	bool Is_Isolated() const
		/* The string is isolated when it is writable. 
		 */
		{
			return raw == NULL || (! Constant(raw) && Deref(head).ref_count == 1);
		}

	bool Isolate();
	/* Make the string an allocated HEAD with REF_COUNT == 1. 

	   Return FALSE on error. 
	*/
#if 0
	void Cut(int n);
	/* Loescht n Zeichen am Ende.  Laenge muss >= n sein.  Der
	   String muss isoliert sein. 
	*/
#endif /* 0 */ 
	String Sub(int begin, 
		   int length) 
		const;

	const String &operator= (const String &);
	const String &operator= (const char *);
	const String &operator= (char c)
		{
			return This= String(c);
		}

	String operator+(const String &) const;
	String operator+(const char *) const;
	String operator+(char c) const
		{
			return This + String(c);
		}

	String &operator += (const String &);
	String &operator += (const char *);
	String &operator += (char c)
		{
			return This += String(c);
		}

	bool operator==(const String &) const;
	bool operator == (const char *string) const
		{
			if (string == NULL)
				return raw == NULL;
			if (raw == NULL)
  				return string == NULL || string[0] == '\0'; 
			return 0 == strcmp(This(), string); 
		}
	bool operator==(char c) const
		{
			return *This == 1 && This[0] == c; 
		}

	bool operator != (const String &str) const 
		{
			return !operator==(str);
		}
	bool operator != (const char *p) const
		{
			return ! operator == (p);
		}
	bool operator != (char c) const
		{
			return !operator==(c);
		}

	bool Begins_With(String s) const
		{
			return (! s.Is()      ||
				(Is() && 0 == strncmp(This(), s(), s.Get_Length()))); 
		}

	bool operator < (const String &string) const
		{
			return strcmp(This(), string()) < 0;
		}
	bool operator > (const String &string) const
		{
			return strcmp(string(), This()) < 0; 
		}

	void *Get_Raw()
		{
			return raw; 
		}

	void *Detach()
		{
			if (raw && ! Constant(raw))
				Assert (Deref(head).ref_count >= 1); 
			void *result= raw;
			raw= NULL;
			return result;
		}

	static void Free(void *raw);

private:

	union
	{
	public:

		void *raw;
		Head *head;
		const char *constant;
	};

	bool Do_Assign(const char *string);
	/* Does not deallocate current HEAD (if any).  STRING must not
	   be NULL.  On failure, set RAW to NULL. 
	*/

	static void Free_Head(Head *head);
	/* Free the string without setting HEAD.  Must be non-empty
	   and non-constant. 
	*/
};

String operator + (char c, String s);

Inline
String
operator + (const char *string_1, 
	    String string_2)
{
	return String(string_1) + string_2;
}

#endif /* ! STRING_INCLUDED */
