#ifndef VECTOR_INCLUDED
#define VECTOR_INCLUDED

/* An array of variable size. 
   
   A vector holds N objects of the same type.  The number of objects
   can be changed dynamically.  There cannot be vectors of objects
   that are not moveable, as defined by the MOVEABLE module.

   The objects in the vector are moved in memory.  Their addresses are
   not constant. 
 */

#include "mem.hh"
#include "assert.hh"
#include "object.hh"
#include "moveable.hh"
#include "this.hh"
#include "pointer.hh"
#include "dbg.hh"
#include "type_name.hh"

template <class Type>
class Vector
	:  public Pointed
{
public:

	Vector()
		:  pointer(NULL),
	           size(0),
		   place(0) 
		{
		}

	Vector(const Vector &vector)
		:  Pointed(),
		   pointer(NULL),
		   size(0),
		   place(0)
		{
			This= vector;
		}

	~Vector() 
		{
			for (int i= 0; i < size; ++i)
				Destructor(Type, pointer + i);
			Mem::Free(pointer); 
			if (pointer != NULL)
				Dbg("%12p %12p <-- %12p %s", 
				    this, 
				    (void *)NULL,
				    pointer,
				    Type_Name(Type));
		}

	int Get_Size() const 
		{
			return size; 
		}

	int operator * () const
		{
			return Get_Size();
		}

	bool Is() const
		/* Wether size is greater than 0.
		 */
		{
			return size != 0;
		}


	const Type *operator () () const
		{
			return pointer;
		}

	Type *operator () () 
		{
			return pointer; 
		}
	Type *operator () (int i)
		{
			int index= i < 0 ? i + size : i;
			Assert (index >= 0 && index < size);
			return pointer + index;
		}

	const Type *operator () (int i) const
		{
			int index= i < 0 ? i + size : i;
			Assert (index >= 0 && index < size);
			return pointer + index;
		}

	Type &operator [] (int i)
		{
			Assert(i >= - size && i < size);
			int index= i < 0 ? i + size : i;
			Assert (index >= 0 && index < size);
			return pointer[index];
		}

	const Type &operator [] (int i) const
		{
			Assert(i >= - size && i < size);
			int index= i < 0 ? i + size : i;
			Assert (index >= 0 && index < size);
			return pointer[index];
		}

	void Free()
		{
			for (int i= 0; i < size; ++i)
				Destructor(Type, pointer + i);
			Mem::Free(pointer);
			Dbg("%12p %12p <-- %12p %s", 
			    this, 
			    (void *)NULL,
			    pointer,
			    Type_Name(Type));
			pointer= NULL;
			size= 0;
			place= 0;
		}

	bool Set(int n);
	/* Change the size.  Return FALSE on error.  Making
	   smaller always succeeds. 
	*/ 

	void operator = (int new_size)
		/* Change the size.  Lose the return value. 
		 */
		{
			Set(new_size);
		}

	Type *Detach ()
		{
			Type *p= pointer;
			pointer= NULL;
			size= 0;
			place= 0;
			return p;
		}

	void operator = (const Vector <Type> &vector)
		{
			Free();
			if (! Set(*vector))
				return;
			for (int i= 0;  i < size;  ++i)
				This[i]= vector[i];
		}

	bool Append(const Type &t)
		{
			if (! Set(*This + 1))
				return false;
			This[-1]= t;
			return true;
		}

	void operator += (const Type &object)
		{
			Append(object); 
		}

	bool Append(const Vector <Type> &vector)
		{
			int size= *This;
			if (! Set(*This + *vector))
				return false;
			for (int i= 0;  i < *vector;  ++i)
				This[size + i]= vector[i]; 
			return true;
		}
	void Append_Transfer_From(Vector <Type> &vector,
				  int i= 0);
	/* Append objects from VECTOR[i] to the end of VECTOR.  Shrink
	   VECTOR to the size I.
	 */

private:

	Type *pointer; 
	/* Pointer to array of objects. NULL if SIZE is 0.   
	 */
	int size;
	/* Number of objects on array.  If 0, POINTER is NULL. 
	 */
	int place;
	/* The number of objects that fit in *POINTER.  PLACE >= SIZE.   
	 */

	static int Plus(int n)
		/* Criterium for not calling Mem::Realloc() too
		   often.  When calling Mem::Realloc(), reserve at
		   least place for PERCENT more memory. 
		*/
		{
			const int percent= 5;
			return n * (100 + percent) / 100;
		}
};

template <class Type>
bool Vector <Type> ::Set(int n)
{
	Dbg("%p  Set(), pointer= %p", this, pointer);

	if (n < 0)
		n += size;
	Assert (n >= 0);
	if (n <= size)
				/* make smaller */
	{
		for (int i= n;  i < size;  ++i)
			Destructor(Type, pointer + i);
		size= n;
#ifndef NDEBUG
		for (int i= 0;  i < n;  ++i)
			Assert(Moveable(pointer + i));
#endif /* ! NDEBUG */
		Type *p= (Type *)Mem::Re_Alloc(pointer, n * sizeof (Type));
				/* If Mem::Re_Alloc() fails, keep the
				   current size and succeed. 
				*/
		if (p == NULL && n != 0)
			return true;
		Dbg("%12p %12p <-- %12p %s", 
		    this, 
		    p,
		    pointer,
		    Type_Name(Type));
		pointer= p;
		place= n;
		return true;
	}
	else if (n > place)
				/* make bigger than current place */
	{
		int nn= n;
		int plus= Plus(place);
		if (plus > nn)
			nn= plus;
#ifndef NDEBUG
		for (int i= 0;  i < size;  ++i)
			Assert(Moveable(pointer + i));
#endif /* ! NDEBUG */
		Type *p= (Type *)Mem::Re_Alloc(pointer, nn * sizeof(Type));
		if (p == NULL)
			return false;

		Dbg("%12p %12p <-- %12p %s", 
		    this, 
		    p,
		    pointer,
		    Type_Name(Type));
		pointer= p;
		for (int i= size; i < n; ++i)
			Constructor(Type, pointer + i);
		size= n;
		place= nn;
		return true;
	}
	else 
				/* make bigger, but have enough place */
	{
		for (int i= size; i < n; ++i)
			Constructor(Type, pointer + i);
		size= n;
		return true;
	}
}

template <class Type>
void Vector <Type> ::
Append_Transfer_From(Vector <Type> &vector,
		     int i)
{
	Assert (i >= 0);
	Assert (i < vector.size);

#ifndef NDEBUG
	for (int j= i;  j < vector.size;  ++j)
		Assert(Moveable(vector.pointer + j));
#endif
	Type *p= (Type *)Mem::Re_Alloc
		(pointer, 
		 (size + vector.size - i) * sizeof(Type));
	if (p == NULL)
		return;
	pointer= p;
	memcpy(pointer + size,
	       vector.pointer + i,
	       (vector.size - i) * sizeof (Type));
	size += vector.size - i;
	place= size;
	p= (Type *)Mem::Re_Alloc(vector.pointer, i * sizeof (Type));
	if (p != NULL)
	{
		vector.pointer= p;
		vector.place= i;
	}
	else if (i == 0)
	{
		vector.pointer= NULL; 
		vector.place= 0; 
	}
	vector.size= i;
}

#endif /* ! VECTOR_INCLUDED */
